@@ -1,7 +1,7 @@
 /*
     Qalculate    
 
-    Copyright (C) 2003-2007, 2008, 2016  Hanna Knutsson (hanna.knutsson@protonmail.com)
+    Copyright (C) 2003-2007, 2008, 2016-2018  Hanna Knutsson (hanna.knutsson@protonmail.com)
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -125,6 +125,8 @@ struct PlotDataParameters {
 	bool yaxis2;
 	/// Use scale on second x-axis
 	bool xaxis2;
+	/// Check if data is continuous
+	bool test_continuous;
 	PlotDataParameters();
 };
 
@@ -208,6 +210,8 @@ class Calculator {
 
 	int ianglemode;
 	int i_precision;
+	bool b_interval;
+	int i_stop_interval;
 	char vbuffer[200];
 	vector<void*> ufvl;
 	vector<char> ufvl_t;
@@ -219,15 +223,16 @@ class Calculator {
 	
 	class Calculator_p *priv;
 	
-	vector<string> signs;	
+	vector<string> signs;
 	vector<string> real_signs;
-	vector<string> default_signs;	
-	vector<string> default_real_signs;	
+	vector<string> default_signs;
+	vector<string> default_real_signs;
 	char *saved_locale;
 	int disable_errors_ref;
 	vector<int> stopped_errors_count;
 	vector<int> stopped_warnings_count;
 	vector<int> stopped_messages_count;
+	vector<vector<CalculatorMessage> > stopped_messages;
 
 	Thread *calculate_thread;
 
@@ -251,6 +256,8 @@ class Calculator {
 	vector<MathFunction*> deleted_functions;
 	vector<Unit*> deleted_units;
 	
+	bool b_var_units;
+	
 	bool b_save_called;
 	
 	int i_timeout;
@@ -268,41 +275,44 @@ class Calculator {
 	
   public:
 
-	KnownVariable *v_pi, *v_e, *v_euler, *v_catalan, *v_i, *v_inf, *v_pinf, *v_minf, *v_undef, *v_precision;
+	KnownVariable *v_pi, *v_e, *v_euler, *v_catalan, *v_i, *v_pinf, *v_minf, *v_undef, *v_precision;
+	KnownVariable *v_percent, *v_permille, *v_permyriad;
+	KnownVariable *v_today, *v_yesterday, *v_tomorrow, *v_now;
 	UnknownVariable *v_x, *v_y, *v_z;
 	MathFunction *f_vector, *f_sort, *f_rank, *f_limits, *f_component, *f_dimension, *f_merge_vectors;
 	MathFunction *f_matrix, *f_matrix_to_vector, *f_area, *f_rows, *f_columns, *f_row, *f_column, *f_elements, *f_element, *f_transpose, *f_identity, *f_determinant, *f_permanent, *f_adjoint, *f_cofactor, *f_inverse; 
 	MathFunction *f_factorial, *f_factorial2, *f_multifactorial, *f_binomial;
 	MathFunction *f_xor, *f_bitxor, *f_even, *f_odd, *f_shift;
-	MathFunction *f_abs, *f_gcd, *f_lcm, *f_signum, *f_round, *f_floor, *f_ceil, *f_trunc, *f_int, *f_frac, *f_rem, *f_mod;	
+	MathFunction *f_abs, *f_gcd, *f_lcm, *f_signum, *f_heaviside, *f_dirac, *f_round, *f_floor, *f_ceil, *f_trunc, *f_int, *f_frac, *f_rem, *f_mod;	
 	MathFunction *f_polynomial_unit, *f_polynomial_primpart, *f_polynomial_content, *f_coeff, *f_lcoeff, *f_tcoeff, *f_degree, *f_ldegree;
-	MathFunction *f_re, *f_im, *f_arg, *f_numerator, *f_denominator;
+	MathFunction *f_re, *f_im, *f_arg, *f_numerator, *f_denominator, *f_interval;
   	MathFunction *f_sqrt, *f_cbrt, *f_root, *f_sq;
 	MathFunction *f_exp;
 	MathFunction *f_ln, *f_logn;
 	MathFunction *f_lambert_w;
-	MathFunction *f_sin, *f_cos, *f_tan, *f_asin, *f_acos, *f_atan, *f_sinh, *f_cosh, *f_tanh, *f_asinh, *f_acosh, *f_atanh, *f_atan2, *f_radians_to_default_angle_unit;
+	MathFunction *f_sin, *f_cos, *f_tan, *f_asin, *f_acos, *f_atan, *f_sinh, *f_cosh, *f_tanh, *f_asinh, *f_acosh, *f_atanh, *f_atan2, *f_sinc, *f_radians_to_default_angle_unit;
 	MathFunction *f_zeta, *f_gamma, *f_digamma, *f_beta, *f_airy, *f_besselj, *f_bessely, *f_erf, *f_erfc;
 	MathFunction *f_total, *f_percentile, *f_min, *f_max, *f_mode, *f_rand;
-	MathFunction *f_isodate, *f_localdate, *f_timestamp, *f_stamptodate, *f_days, *f_yearfrac, *f_week, *f_weekday, *f_month, *f_day, *f_year, *f_yearday, *f_time, *f_add_days, *f_add_months, *f_add_years;
+	MathFunction *f_date, *f_timevalue, *f_timestamp, *f_stamptodate, *f_days, *f_yearfrac, *f_week, *f_weekday, *f_month, *f_day, *f_year, *f_yearday, *f_time, *f_add_days, *f_add_months, *f_add_years;
 	MathFunction *f_bin, *f_oct, *f_hex, *f_base, *f_roman;
 	MathFunction *f_ascii, *f_char;
 	MathFunction *f_length, *f_concatenate;
 	MathFunction *f_replace, *f_stripunits;
 	MathFunction *f_genvector, *f_for, *f_sum, *f_product, *f_process, *f_process_matrix, *f_csum, *f_if, *f_is_number, *f_is_real, *f_is_rational, *f_is_integer, *f_represents_number, *f_represents_real, *f_represents_rational, *f_represents_integer, *f_function, *f_select;
-	MathFunction *f_diff, *f_integrate, *f_solve, *f_multisolve;
+	MathFunction *f_diff, *f_integrate, *f_solve, *f_multisolve, *f_dsolve;
+	MathFunction *f_li, *f_Li, *f_Ei, *f_Si, *f_Ci, *f_Shi, *f_Chi, *f_igamma;
 	MathFunction *f_error, *f_warning, *f_message, *f_save, *f_load, *f_export, *f_title;
 	MathFunction *f_register, *f_stack;
 	MathFunction *f_plot;
-	MathFunction *f_uncertainty;
 	
-	Unit *u_rad, *u_gra, *u_deg, *u_euro;
+	Unit *u_rad, *u_gra, *u_deg, *u_euro, *u_btc, *u_second, *u_minute, *u_hour, *u_year, *u_month, *u_day;
 	DecimalPrefix *decimal_null_prefix;
 	BinaryPrefix *binary_null_prefix;
 
   	bool place_currency_code_before, place_currency_sign_before;
 	bool place_currency_code_before_negative, place_currency_sign_before_negative;
 	bool default_dot_as_separator;
+	string local_digit_group_separator, local_digit_group_format;
   
   	bool b_busy;
 	string expression_to_calculate;
@@ -585,7 +595,7 @@ class Calculator {
 	* @param str The expression to localize.
 	* @returns A localized expression.
 	*/
-	string localizeExpression(string str) const;
+	string localizeExpression(string str, const ParseOptions &po = default_parse_options) const;
 	/** Returns an unlocalized expressions. Affects decimal signs and argument separators.
 	*
 	* @param str The expression to unlocalize.
@@ -602,7 +612,7 @@ class Calculator {
 	bool separateToExpression(string &str, string &to_str, const EvaluationOptions &eo, bool keep_modifiers = false) const;	
 	bool hasToExpression(const string &str) const;
 
-	void parseSigns(string &str) const;
+	void parseSigns(string &str, bool convert_to_internal_representation = false) const;
 	/** Parse an expression and place in a MathStructure object.
 	*
 	* @param str Expression
@@ -639,6 +649,7 @@ class Calculator {
 	* @returns Converted value.
 	*/
 	MathStructure convert(const MathStructure &mstruct, Unit *to_unit, const EvaluationOptions &eo = default_evaluation_options, bool always_convert = true, bool convert_to_mixed_units = true);
+	MathStructure convert(const MathStructure &mstruct, KnownVariable *to_var, const EvaluationOptions &eo = default_evaluation_options);
 	MathStructure convert(double value, Unit *from_unit, Unit *to_unit, const EvaluationOptions &eo = default_evaluation_options);
 	MathStructure convert(string str, Unit *from_unit, Unit *to_unit, int milliseconds, const EvaluationOptions &eo = default_evaluation_options);
 	/** Depecated: use convert() */
@@ -851,6 +862,9 @@ class Calculator {
 	/** Adds builtin units. Called automatically when the calculator is created. */
 	void addBuiltinUnits();
 	//@}
+	
+	void setVariableUnitsEnabled(bool enable_variable_units = true);
+	bool variableUnitsEnabled() const;
 
 	/** @name Functions for testing validity of functions, variable and unit names. */
 	//@{
@@ -910,7 +924,9 @@ class Calculator {
 	CalculatorMessage *nextMessage();
 	bool showArgumentErrors() const;
 	void beginTemporaryStopMessages();
-	int endTemporaryStopMessages(int *message_count = NULL, int *warning_count = NULL);	
+	int endTemporaryStopMessages(int *message_count = NULL, int *warning_count = NULL, int release_messages_if_no_equal_or_greater_than_message_type = -1);
+	void endTemporaryStopMessages(bool release_messages, vector<CalculatorMessage> *blocked_messages = NULL);
+	void addMessages(vector<CalculatorMessage> *message_vector);
 	//@}
 
 	/** @name Functions for loading and saving definitions (variables, functions, units, etc.). */
@@ -1006,17 +1022,19 @@ class Calculator {
 	*/
 	bool loadExchangeRates();
 	/** Name of the exchange rates file on local disc.
+	* Multiple exchange rates sources might be used. Iterate over these, using the index parameter, until an empty string is returned.
 	*
+	* @param index The index (starting at one) of the exchange rate source
 	* @returns name of local exchange rates file.
 	*/
-	string getExchangeRatesFileName();
-	string getExchangeRatesFileName2();
+	string getExchangeRatesFileName(int index = 1);
 	/** Url of the exchange rates file on the Internet.
+	* Multiple exchange rates sources might be used. Iterate over these, using the index parameter, until an empty string is returned.
 	*
+	* @param index The index (starting at one) of the exchange rate source
 	* @returns Url of exchange rates file.
 	*/
-	string getExchangeRatesUrl();
-	string getExchangeRatesUrl2();
+	string getExchangeRatesUrl(int index = 1);
 	/** Modification time of the exchange rates file.
 	*
 	* @returns Returns exchange rates modification time.
@@ -1076,6 +1094,16 @@ class Calculator {
 	/** Returns default precision for approximate calculations.
 	*/
 	int getPrecision() const;
+	/** Set if interval arithmetic should be used for approximate calculations.
+	*
+	* @param use_interval_arithmetic Set true to activate, or false to deactivate, interval arithmetic.
+	*/
+	void useIntervalArithmetic(bool use_interval_arithmetic = true);
+	/** Returns true if interval arithmetic are activated.
+	*/
+	bool usesIntervalArithmetic() const;
+	void beginTemporaryStopIntervalArithmetic();
+	void endTemporaryStopIntervalArithmetic();
 	//@}
 
 	/** @name Functions for localization */
@@ -1084,12 +1112,12 @@ class Calculator {
 	*/
 	const string &getDecimalPoint() const;
 	/** Returns the preferred comma character for separating arguments.*/
-	const string &getComma() const;	
+	const string &getComma() const;
 	/** Sets argument separator and decimal sign from the current locale. Mainly for internal use. */
 	void setLocale();
 	void useDecimalComma();
 	/** Use point as decimal separator. 
-	* To use comma as an ignored separator in numbers, must be invoked with comma_as_separator = true, before using ParseOptions::comma_as_separator.
+	* To use comma as an ignored separator in numbers, must be invoked with comma_as_separator = true, to change the default function argument separator to semicolon, in addition to using ParseOptions::comma_as_separator.
 	*/	
 	void useDecimalPoint(bool comma_as_separator = false);
 	/** Resets argument separator and decimal sign. Mainly for internal use. */
@@ -1141,43 +1169,4 @@ class Calculator {
 		
 };
 
-class QalculateDate {
-	protected:
-		long int i_year;
-		long int i_month;
-		long int i_day;
-	public:
-		QalculateDate();
-		QalculateDate(long int initialyear, int initialmonth, int initialday);
-		QalculateDate(long int initialtimestamp);
-		QalculateDate(string date_string);
-		bool operator > (const QalculateDate &date2) const;
-		bool operator < (const QalculateDate &date2) const;
-		bool operator >= (const QalculateDate &date2) const;
-		bool operator <= (const QalculateDate &date2) const;
-		bool operator != (const QalculateDate &date2) const;
-		bool operator == (const QalculateDate &date2) const;
-		bool isFutureDate() const;
-		bool isPastDate() const;
-		void setToCurrentDate();
-		bool set(long int newyear, int newmonth, int newday);
-		bool set(long int newtimestamp);
-		bool set(string date_string);
-		string toISOString() const;
-		string toLocalString() const;
-		long int year() const;
-		long int month() const;
-		long int day() const;
-		bool addDays(long int days);
-		bool addMonths(long int months);
-		bool addYears(long int years);
-		int weekday() const;
-		int week(bool start_sunday = false) const;
-		int yearday() const;
-		Number timestamp() const;
-		Number daysTo(const QalculateDate &date, int basis = 1, bool date_func = true) const;
-		Number yearsTo(const QalculateDate &date, int basis = 1, bool date_func = true) const;
-};
-
-
 #endif
